/*
 * Copyright (c) 2017 Gregory Penn.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU Affero General
 *     Public License as published by the Free Software Foundation,
 *     either version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU Affero General Public License for more
 *     details.
 *
 *     You should have received a copy of the GNU Affero General
 *     Public License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

#ifndef GDAL_SMARTPTR_HPP
#define GDAL_SMARTPTR_HPP

#include <ogr_feature.h>
#include <functional>
#include <utility>
#include <memory>
#include <gdal/ref_manager.hpp>

namespace gdal
{

  template < typename T >
  class smartptr
  {
    template < typename U >
    friend class smartptr;

    friend class ref_manager;

    using self = smartptr<T>;

    using deleter_type = std::function<void(T*)>;

  public:
    smartptr () = default;

    smartptr (T* p
              , std::function<void()> deleter
              , std::function<void()> increment
              , std::function<void()> decrement
              , std::function<int()>  use_count)
      : p_(p)
      , ref_manager_(deleter, increment, decrement, use_count)
      { owner_ = ref_manager_; }

    template < typename U >
    smartptr (T* p
              , smartptr<U> owner
              , std::function<void()> deleter
              , std::function<void()> increment
              , std::function<void()> decrement
              , std::function<int()>  use_count)
      : p_(p)
      , ref_manager_(deleter, increment, decrement, use_count)
      , owner_(owner.ref_manager_)
      {}


  public:
    auto     operator ->   () const { return get(); }
    auto&    operator *    () const { return *get(); }
    explicit operator bool () const { return static_cast<bool>(p_); }

    auto use_count () const { return ref_manager_.use_count(); }

    auto get () const { assert(p_); return p_; }

  private:
    T* p_ = nullptr;
    ref_manager ref_manager_;
    ref_manager owner_;

  };

  template <>
  class smartptr < OGRFeature >
  {
    template < typename U >
    friend class smartptr;

    friend class ref_manager;

    using self = smartptr<OGRFeature>;
    using gdal_ptr = std::shared_ptr<OGRFeature>;

  public:
    smartptr () = default;

    template < typename U >
    smartptr (std::function<OGRFeature*()> f
              , smartptr<U> owner
      )
      : p_(f(), &OGRFeature::DestroyFeature)
      {}

  public:
    auto     operator ->   () const { return p_; }
    auto&    operator *    () const { return *get(); }
    explicit operator bool () const { return static_cast<bool>(p_); }

    auto use_count () const { return p_.use_count(); }

    gdal_ptr const& get () const { assert(p_); return p_; }

  private:
    gdal_ptr p_;
    ref_manager owner_;

  };

} // namespace gdal

#endif
