/*
 * Copyright (c) 2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU Affero General
 *     Public License as published by the Free Software Foundation,
 *     either version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU Affero General Public License for more
 *     details.
 *
 *     You should have received a copy of the GNU Affero General
 *     Public License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file

#ifndef GRAPH_MODEL_MCMC_RANDOM_VARIABLE_PROXY_HPP
#define GRAPH_MODEL_MCMC_RANDOM_VARIABLE_PROXY_HPP 1

#include <functional>
#include <graph_model/incidence_graph.hpp>
#include <graph_model/vertex_and_edge_list_graph.hpp>
#include <logging/channel_log.hpp>
#include <logging/global_log.hpp>

namespace graph_model
{

  namespace MCMC
  {

    class random_variable_proxy
    {
      using function_type = std::function<void()>;
      using graph_callback_type = std::function<void(vertex_and_edge_list_graph&)>;
      using log_type = logging::channel_log;

    public:
      random_variable_proxy () = default;

      template < typename RandomVariable, typename Initialization >
      random_variable_proxy (RandomVariable& rv, Initialization initialization)
	: use_current_variate_(initialize_use_current_variate(rv))
	, use_proposed_variate_(initialize_use_proposed_variate(rv))
	, accept_(initialize_accept(rv))
	, reject_(initialize_reject(rv))
	, commit_(initialize_commit(rv))
	, graph_callback_(initialize_graph_callback(rv))
      { initialize(rv,initialization); }

      void use_current_variate () { use_current_variate_(); }
      void use_proposed_variate () { use_proposed_variate_(); }
      void accept () { accept_(); }
      void reject () { reject_(); }
      void commit () { commit_(); }
      void graph_callback (vertex_and_edge_list_graph& graph)
      {
	using namespace boost::log::trivial;
	GLOBAL_LOG(log_,trace) << "graph_callback(vertex_and_edge_list_graph&) [" << this << "]";
	graph_callback_(graph);
	GLOBAL_LOG(log_,trace) << "graph_callback(vertex_and_edge_list_graph&) [" << this << "] completed";
      }

    private:
      template < typename RandomVariable, typename Initialization >
      auto initialize (RandomVariable& rv, Initialization initialization)
      {
	initialization(rv);
      }

      template < typename RandomVariable >
      static auto initialize_use_current_variate (RandomVariable& rv) { return [&rv](){ rv.use_current_variate(); }; }

      template < typename RandomVariable >
      static auto initialize_use_proposed_variate (RandomVariable& rv) { return [&rv](){ rv.use_proposed_variate(); }; }

      template < typename RandomVariable >
      static auto initialize_accept (RandomVariable& rv) { return [&rv](){ rv.accept(); }; }

      template < typename RandomVariable >
      static auto initialize_reject (RandomVariable& rv) { return [&rv](){ rv.reject(); }; }

      template < typename RandomVariable >
      static auto initialize_commit (RandomVariable& rv) { return [&rv](){ rv.commit(); }; }

      template < typename RandomVariable >
      static auto initialize_graph_callback (RandomVariable& rv)
      {
	return [&rv](auto& graph){ graph.add_connected_components(make_incidence_graph(),rv); };
      }

    private:
      function_type use_current_variate_;
      function_type use_proposed_variate_;
      function_type accept_;
      function_type reject_;
      function_type commit_;
      graph_callback_type graph_callback_;
      mutable log_type log_ = "graph_model::MCMC::random_variable_proxy";
    };

    template < typename RandomVariable, typename Initialization >
    auto make_random_variable_proxy (RandomVariable& rv, Initialization initialization)
    {
      return random_variable_proxy(rv,initialization);
    }

  } // namespace MCMC

} // namespace graph_model

#endif
