/*
 * Copyright (c) 2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU Affero General
 *     Public License as published by the Free Software Foundation,
 *     either version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU Affero General Public License for more
 *     details.
 *
 *     You should have received a copy of the GNU Affero General
 *     Public License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file

#ifndef GRAPH_MODEL_DISTRIBUTION_BERNOULLI_BERNOULLI_STRATEGY_HPP
#define GRAPH_MODEL_DISTRIBUTION_BERNOULLI_BERNOULLI_STRATEGY_HPP 1

#include <type_traits>
#include <boost/log/sources/severity_logger.hpp>
#include <boost/log/trivial.hpp>
#include <boost/math/distributions/bernoulli.hpp>
#include <boost/random/bernoulli_distribution.hpp>
#include <graph_model/distribution/distribution_policy.hpp>
#include <graph_model/distribution/parameter/probability.hpp>
#include <graph_model/distribution/parameter/parameter_pack.hpp>
#include <graph_model/distribution/parameter/variate.hpp>
#include <graph_model/random/distribution.hpp>
#include <graph_model/traits/real.hpp>
#include <logging/channel_log.hpp>
#include <logging/global_log.hpp>
#include <sample_space/is_possible.hpp>

namespace graph_model
{

  namespace distribution
  {

    namespace bernoulli
    {

      /// \brief Generate a Bernoulli distribution
      ///
      /// This policy class models the Distribution Generator concept,
      /// which generates a distribution from its parameters.  A
      /// distribution models the corresponding concept in the Boost
      /// Math library.  Specifically, it can be evaluated by the \c
      /// pdf() function (and possibly the \c cdf() function if it is
      /// a scalar distribution).
      ///
      /// This policy generates a Bernoulli distribution from the Boost
      /// Math library.
      ///
      struct distribution_policy
      {
	template < typename Variate, typename P >
	static auto apply (P p)
	{
	  using real_type
	    = std::common_type_t<
	      traits::real_t<Variate>
	    , traits::real_t<P>
	    >;
	  using bernoulli_distribution
	    = boost::math::bernoulli_distribution<real_type>;
	  return bernoulli_distribution(p);
	}
      };

      /// \brief Generate a Bernoulli random number generator
      ///
      /// This policy class models the RNG Generator concept, which
      /// generates a random number distribution from its parameters.
      /// A random number distribution models the corresponding
      /// concept in the Boost Random library.  Specifically, it can
      /// be evaluated to obtain a sample from the distribution.
      ///
      /// This policy generates a Bernoulli distribution from the Boost
      /// Random library.
      ///
      struct generator_policy
      {
	template < typename P >
	static auto& apply ()
	{
	  using real_type = traits::real_t<P>;
	  using bernoulli_distribution
	    = random::distribution< boost::bernoulli_distribution<real_type> >;
	  return bernoulli_distribution::apply();
	}
      };

      /// \brief Strategy for Bernoulli distributions
      /// \tparam ParameterPack The parameter pack containing the parameters
      ///
      /// A Bernoulli distribution is a distribution across a variate
      /// that is constructible from a boolean value.  Generally, the
      /// implementation simply delegates to the appropriate
      /// components of the Boost Math or Random libraries.
      ///
      template < typename ParameterPack >
      class bernoulli_strategy
      {
	using policy_type = distribution::distribution_policy<distribution_policy,generator_policy>;
	using log_type = logging::channel_log;

      public:
	bernoulli_strategy () = default;
	template < typename Tags, typename Defaults, typename Tagged, typename Untagged >
	explicit bernoulli_strategy (parameter::parameter_pack<Tags,Defaults,Tagged,Untagged> pack)
	  : pack_(pack)
	{}

      public:
	template < typename ... T >
	auto variate (T&& ... t) const
	{ return pack_.template value<parameter::tag::variate>(std::forward<T>(t)...); }

	template < typename ... T >
	auto p (T&& ... t) const
	{ return pack_.template value<parameter::tag::probability>(std::forward<T>(t)...); }

      public:
	template < typename ... T >
	auto sample (T&& ... t) const
	{
	  using namespace boost::log::trivial;

	  auto p = this->p(std::forward<T>(t)...);
	  auto result = sample(p);

	  GLOBAL_LOG(log_,trace)
	    << "p=" << p << ", "
	    << "sample=" << result;

	  return result;
	}

	template < typename ... T >
	auto pdf (T&& ... t) const
	{
	  using namespace boost::log::trivial;

	  auto variate = this-> variate(std::forward<T>(t)...);
	  auto p = this->p(std::forward<T>(t)...);
	  auto result = pdf(variate,p);

	  GLOBAL_LOG(log_,trace)
	    << "variate=" << variate << ", "
	    << "p=" << p << ", "
	    << "pdf=" << result;

	  return result;
	}

      private:
	template < typename P >
	auto sample (P p) const
	{
	  using sample_space::is_possible;
	  assert(is_possible(p));
	  return policy_type::sample(p);
	}

	template < typename Variate, typename P >
	auto pdf (Variate variate, P p) const
	{
	  using sample_space::is_possible;
	  return
	    0 <= p && p <= 1
	    && is_possible(p)
	    ? policy_type::eval_pdf(variate,p)
	    : 0;
	}

      private:
	ParameterPack pack_;
	mutable log_type log_ = "graph_model::distribution::bernoulli";
      };

      template < typename ParameterPack >
      auto make_bernoulli_strategy (ParameterPack parameter_pack)
      {
	using bernoulli_strategy_type = bernoulli_strategy<ParameterPack>;
	return bernoulli_strategy_type(parameter_pack);
      };

    } // namespace bernoulli

  } // namespace distribution

} // namespace graph_model

#endif
