/*
 * Copyright (c) 2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU Affero General
 *     Public License as published by the Free Software Foundation,
 *     either version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU Affero General Public License for more
 *     details.
 *
 *     You should have received a copy of the GNU Affero General
 *     Public License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file

#ifndef GRAPH_MODEL_DISTRIBUTION_DIRICHLET_DIRICHLET_STRATEGY_HPP
#define GRAPH_MODEL_DISTRIBUTION_DIRICHLET_DIRICHLET_STRATEGY_HPP 1

#include <type_traits>
#include <boost/log/sources/severity_logger.hpp>
#include <boost/log/trivial.hpp>

#include <graph_model/distribution/dirichlet/tags.hpp>
#include <graph_model/distribution/distribution_policy.hpp>
#include <graph_model/random/distribution.hpp>
#include <graph_model/vertex/parameter_tags.hpp>
#include <graph_model/vertex/tagged_parameter.hpp>
#include <sample_space/assert_is_possible.hpp>
#include <sample_space/is_possible.hpp>

#include <graph_model/distribution/parameter/alpha.hpp>
#include <graph_model/distribution/parameter/mean.hpp>
#include <graph_model/distribution/parameter/parameter_pack.hpp>
#include <graph_model/distribution/parameter/precision.hpp>
#include <graph_model/distribution/parameter/variate.hpp>

#include <sample_space/math/dirichlet_distribution.hpp>
#include <sample_space/random/dirichlet_distribution.hpp>

#include <logging/channel_log.hpp>
#include <logging/global_log.hpp>

namespace graph_model
{

  namespace distribution
  {

    namespace dirichlet
    {

      /// \brief Generate a dirichlet distribution
      ///
      /// This policy class models the Distribution Generator concept,
      /// which generates a distribution from its parameters.  A
      /// distribution models the corresponding concept in the Boost
      /// Math library.  Specifically, it can be evaluated by the \c
      /// pdf() function (and possibly the \c cdf() function if it is
      /// a scalar distribution).
      ///
      /// This policy generates a dirichlet distribution, which models
      /// the Random Distribution concept of the Boost Random library.
      ///
      struct distribution_policy
      {
	template < typename Variate, typename Alpha >
	static auto apply (Alpha alpha)
	{
	  using dirichlet_distribution = sample_space::math::dirichlet_distribution<Alpha>;
	  return dirichlet_distribution(alpha);
	}
      };

      /// \brief Generate a dirichlet random number generator
      ///
      /// This policy class models the Random Distribution concept
      /// from the Boost Random library.
      ///
      struct generator_policy
      {
	template < typename Alpha >
	static auto& apply ()
	{
	  using dirichlet_distribution_type = sample_space::random::dirichlet_distribution<Alpha>;
	  static dirichlet_distribution_type dirichlet_distribution;
	  return dirichlet_distribution;
	}
      };

      /// \brief Translate the parameter pack into alpha
      ///
      /// The parameters of the dirichlet distribution may be specified in
      /// one of two ways:
      ///
      /// - alpha and dirichlet
      /// - probability and precision
      ///
      /// However, the underlying distribution and generator are
      /// parameterized with alpha and dirichlet.  Specializations of this
      /// class facilitate translation of a parameter pack holding
      /// either set of parameters into alpha.
      ///
      template < typename Tag >
      class alpha;

      template <>
      struct alpha < tag::alpha >
      {
	template < typename ParameterPack, typename ... T >
	static auto apply (ParameterPack const& pack, T&& ... t)
	{ return pack.template value<parameter::tag::alpha>(std::forward<T>(t)...); }
      };

      template <>
      struct alpha < tag::mean_precision >
      {
	template < typename ParameterPack, typename ... T >
	static auto apply (ParameterPack const& pack, T&& ... t)
	{
	  auto mean = pack.template value<parameter::tag::mean>(std::forward<T>(t)...);
	  auto precision = pack.template value<parameter::tag::precision>(std::forward<T>(t)...);
	  auto alpha = mean * precision;
	  return alpha;
	}
      };

      /// \brief Strategy for dirichlet distributions
      /// \tparam Tag Determine how to interpret the parameters
      /// \tparam ParameterPack The parameters
      ///
      /// A dirichlet distribution is a distribution across a scalar
      /// variate, so this strategy models the Scalar Distribution
      /// concept, implemented as appropriate for a dirichlet distribution.
      /// Generally, the implementation simply delegates to the
      /// appropriate components of the Boost Math or Random
      /// libraries.
      ///
      /// The parameters of a dirichlet distrubtion may be specified in
      /// either of two different ways:
      ///
      /// - as alpha and dirichlet, which is how the underlying Boost
      ///   libraries are parameterized, or
      /// - as the probability and precision.  In this case, alpha
      ///   equals probability * precision.
      ///
      /// Given that different use cases may naturally prefer one of
      /// these over the rest, both are supported.  The \c Tag
      /// template parameter determines how the parameters should be
      /// interpreted.  In either case, both alpha and dirichlet are
      /// calculated before being passed to the underlying Boost
      /// library functions.
      ///
      template < typename Tag, typename ParameterPack >
      class dirichlet_strategy
      {
	using policy_type = distribution::distribution_policy<distribution_policy,generator_policy>;
	using log_type = logging::channel_log;

      public:
	dirichlet_strategy () = default;
	template < typename Tags, typename Defaults, typename Tagged, typename Untagged >
	explicit dirichlet_strategy (parameter::parameter_pack<Tags,Defaults,Tagged,Untagged> pack)
	  : pack_(pack)
	{}

      public:
	template < typename ... T >
	auto variate (T&& ... t) const
	{ return pack_.template value<parameter::tag::variate>(std::forward<T>(t)...); }

	template < typename ... T >
	auto alpha (T&& ... t) const
	{ return dirichlet::alpha<Tag>::apply(pack_,std::forward<T>(t)...); }

	template < typename ... T >
	auto mean (T&& ... t) const
	{
	  auto alpha = this->alpha(std::forward<T>(t)...);
	  auto alpha0 = this->alpha0(std::forward<T>(t)...);
	  return alpha / alpha0;
	}

	template < typename ... T >
	auto precision (T&& ... t) const
	{ return alpha0(std::forward<T>(t)...); }

      private:
	template < typename ... T >
	auto alpha0 (T&& ... t) const
	{
	  auto alpha = this->alpha(std::forward<T>(t)...);
	  return sum(alpha);
	}

      public:
	template < typename ... T >
	auto sample (T&& ... t) const
	{
	  using namespace boost::log::trivial;

	  auto alpha = this->alpha(std::forward<T>(t)...);
	  auto result = sample(alpha);

	  GLOBAL_LOG(log_,trace)
	    << "alpha=" << alpha << ", "
	    << "sample=" << result;

	  return result;
	}
	template < typename ... T >
	auto pdf (T&& ... t) const
	{
	  using namespace boost::log::trivial;
	  
	  auto variate = this->variate(std::forward<T>(t)...);
	  auto alpha = this->alpha(std::forward<T>(t)...);
	  auto result = pdf(variate,alpha);

	  GLOBAL_LOG(log_,trace)
	    << "variate=" << variate << ", "
	    << "alpha=" << alpha << ", "
	    << "pdf=" << result;

	  return result;
	}

      private:
	template < typename Alpha >
	auto sample (Alpha alpha) const
	{
	  using sample_space::assert_is_possible;
	  assert_is_possible(alpha);
	  return policy_type::sample(alpha);
	}

	template < typename Variate, typename Alpha >
	auto pdf (Variate variate, Alpha alpha) const
	{
	  using namespace boost::log::trivial;
	  using sample_space::is_possible;

	  auto result =
	    is_possible(variate)
	    && is_possible(alpha)
	    ? policy_type::eval_pdf(variate,alpha)
	    : 0;

	  GLOBAL_LOG(log_,trace)
	    << "variate=" << variate << ", "
	    << "alpha=" << alpha << ", "
	    << "pdf=" << result;

	  return result;
	}

      private:
	ParameterPack pack_;
	mutable log_type log_ = "graph_model::distribution::dirichlet";
      };

      template < typename Tag, typename ParameterPack >
      auto make_dirichlet_strategy (ParameterPack parameter_pack)
      {
	using dirichlet_strategy_type = dirichlet_strategy<Tag,ParameterPack>;
	return dirichlet_strategy_type(parameter_pack);
      };

    } // namespace dirichlet

  } // namespace distribution

} // namespace graph_model

#endif
