/*
 * Copyright (c) 2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU Affero General
 *     Public License as published by the Free Software Foundation,
 *     either version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU Affero General Public License for more
 *     details.
 *
 *     You should have received a copy of the GNU Affero General
 *     Public License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file

#ifndef GRAPH_MODEL_VERTEX_SQRT_HPP
#define GRAPH_MODEL_VERTEX_SQRT_HPP 1

#include <cmath>
#include <graph_model/vertex/expression_function.hpp>
#include <graph_model/vertex/expression_terminal.hpp>
#include <mpl/enable_constructor.hpp>

namespace graph_model
{

  namespace vertex
  {

    /// \brief Square root function
    /// \tparam Expression The type of expression to evaluate as the argument
    ///
    template < typename Expression >
    class sqrt_function
    {
      using self = sqrt_function;

    public:
      template < typename Expr
		 , typename = mpl::enable_constructor_t<self,Expr>
		 >
      explicit sqrt_function (Expr&& expr)
	: expression_(std::forward<Expr>(expr))
      {}

      template < typename ... T >
      auto operator () (T&& ... t) const
      {
	using std::sqrt;
	return sqrt(expression_(std::forward<T>(t)...));
      }

    private:
      Expression expression_;
    };

    template < typename ... T, typename ... Args >
    auto transform_expression (expression_terminal<sqrt_function<T...>> expr, Args&& ... args)
    { return evaluate_terminal(expr,std::forward<Args>(args)...); }

    template < typename ... T, typename ... Args >
    auto transform_expression (expression_terminal<sqrt_function<T...>&> expr, Args&& ... args)
    { return evaluate_terminal(expr,std::forward<Args>(args)...); }

    /// \brief Construct a square root function
    ///
    /// This function constructs a sqrt_function object from an \c
    /// expression_function.
    ///
    template < typename Wrapper >
    auto sqrt (expression_function<Wrapper>&& function)
    {
      using function_type = expression_function<Wrapper>;
      using sqrt_function_type = sqrt_function<function_type>;
      return make_expression_function(sqrt_function_type(std::move(function)));
    }

    /// \brief Construct a square root function
    /// \tparam Expr The type of argument
    ///
    /// This function constructs an \c expression_function
    /// representing the square root function.
    ///
    template < typename Expr >
    auto sqrt (Expr&& expr)
    { return sqrt(make_expression_function(std::forward<Expr>(expr))); }

    template < typename Expression >
    auto sqrt (sqrt_function<Expression> const& expr) { return expr; }

    template < typename Expression >
    auto sqrt (sqrt_function<Expression>&& expr) { return std::move(expr); }

  } // namespace vertex

} // namespace graph_model

#endif
