/*
 * Copyright (c) 2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU Affero General
 *     Public License as published by the Free Software Foundation,
 *     either version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU Affero General Public License for more
 *     details.
 *
 *     You should have received a copy of the GNU Affero General
 *     Public License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file

#ifndef GRAPH_MODEL_VERTEX_STRATEGY_BASIC_SCALAR_DISTRIBUTION_HPP
#define GRAPH_MODEL_VERTEX_STRATEGY_BASIC_SCALAR_DISTRIBUTION_HPP 1

#include <memory>
#include <graph_model/concept/scalar_distribution_policy_concept.hpp>
#include <graph_model/vertex/strategy/policy_owner_base.hpp>
#include <graph_model/vertex/strategy/scalar_distribution_base.hpp>
#include <graph_model/vertex/trapping_cdf_base.hpp>
#include <graph_model/vertex/trapping_eval_base.hpp>
#include <graph_model/vertex/trapping_pdf_base.hpp>
#include <graph_model/vertex/trapping_sample_base.hpp>

namespace graph_model
{

  namespace vertex
  {

    namespace strategy
    {

      template < typename Policy
		 , typename Variate
		 , typename Probability
		 >
      class basic_scalar_distribution
	: public scalar_distribution_base<Variate,Probability>
	, policy_owner_base<basic_scalar_distribution<Policy,Variate,Probability>,Policy>
      {
	using self = basic_scalar_distribution;
	using base = scalar_distribution_base<Variate,Probability>;
	using policy_owner_base = policy_owner_base<basic_scalar_distribution<Policy,Variate,Probability>,Policy>;

	BOOST_CONCEPT_ASSERT((concept::scalar_distribution_policy_concept<Policy,Variate,Probability>));

      public:
	basic_scalar_distribution () = default;
	explicit basic_scalar_distribution (Policy const& policy) : policy_owner_base(policy) {}
	explicit basic_scalar_distribution (Policy&& policy) : policy_owner_base(std::move(policy)) {}

      private:
	typename base::result_type
	eval_ () const override { return this->pdf(); }
	typename base::variate_type
	sample_ ( ) const override { return this->policy().sample(); }
	typename base::probability_type
	pdf_ () const override { return this->policy().pdf(); }
	typename base::probability_type
	cdf_ () const override { return this->policy().cdf(); }
	typename base::probability_type
	cdf_ (typename base::variate_type variate) const override { return this->policy().cdf(variate); }
      };

      template < typename Variate, typename Probability >
      struct basic_scalar_distribution< void,Variate,Probability >
	: trapping_eval_base< basic_scalar_distribution<void,Variate,Probability>,
	  trapping_pdf_base< basic_scalar_distribution<void,Variate,Probability>,
	  trapping_sample_base< basic_scalar_distribution<void,Variate,Probability>,
	  trapping_cdf_base< basic_scalar_distribution<void,Variate,Probability>,
	  scalar_distribution_base<Variate,Probability
	  > > > > >
      
      {};

      template < typename Variate, typename Probability >
      auto make_basic_scalar_distribution_policy_pointer ()
      {
	using basic_scalar_distribution_type = basic_scalar_distribution<void,Variate,Probability>;
	return std::make_unique<basic_scalar_distribution_type>();
      }

      template < typename Variate, typename Probability, typename Policy >
      auto make_basic_scalar_distribution_policy_pointer (Policy&& policy)
      {
	using policy_type = std::decay_t<Policy>;
	using basic_scalar_distribution_type = basic_scalar_distribution<policy_type,Variate,Probability>;
	return std::make_unique<basic_scalar_distribution_type>(std::forward<Policy>(policy));
      }

      template < typename Variate, typename Probability >
      struct basic_scalar_distribution_factory
      {
	using pointer = std::unique_ptr<scalar_distribution_base<Variate,Probability>>;

	static pointer apply ()
	{ return make_basic_scalar_distribution_policy_pointer<Variate,Probability>(); }

	template < typename Policy >
	static pointer apply (Policy&& policy)
	{ return make_basic_scalar_distribution_policy_pointer<Variate,Probability>(std::forward<Policy>(policy)); }
      };

    } // namespace strategy

  } // namespace vertex

} // namespace graph_model

#endif
