/*
 * Copyright (c) 2016,2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU Affero General
 *     Public License as published by the Free Software Foundation,
 *     either version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU Affero General Public License for more
 *     details.
 *
 *     You should have received a copy of the GNU Affero General
 *     Public License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file

#define BOOST_TEST_MODULE observer_ptr

#ifdef INCLUDE_UNIT_TEST_HEADERS
#include <boost/test/included/unit_test.hpp>
#else
#include <boost/test/unit_test.hpp>
#endif

#define XXX_CANNOT_CONSTRUCT_MUTABLE_POINTER_FROM_CONST 0

#include "observer_ptr.hpp"
#include <type_traits>
#include <boost/mpl/vector.hpp>

struct mutable_this {};
struct const_this {};

struct C
{
  mutable_this f () { return mutable_this(); }
  const_this f () const { return const_this(); }
};

template < typename T, typename U = void >
struct test
{
  static void default_construct () { default_construct(T()); }
  template < typename V >
  static void construct (V&& v) { T t(std::forward<V>(v)); assert(t); }
  template < typename V >
  static void initialize (V&& v) { T t = { std::forward<V>(v) }; assert(t); }
  template < typename V >
  static void copy_construct (V&& v) { U u(std::forward<V>(v)); T t(u); assert(t); assert(u); }
  template < typename V >
  static void move_construct (V&& v) { U u(std::forward<V>(v)); T t(std::move(u)); assert(t); }
  template < typename V >
  static void copy_assign (V&& v) { T t(std::forward<V>(v)); U u(std::forward<V>(v)); t = u; assert(t); assert(u); }
  template < typename V >
  static void move_assign (V&& v) { T t(std::forward<V>(v)); U u(std::forward<V>(v)); t = std::move(u); assert(t); }
  template < typename V >
  static void test_make_observer (V&& v) { T t = ::make_observer(std::forward<V>(v)); assert(t); }

  static void test_swap () { using std::swap; T t; U u; swap(t,u); }
  static void dereference_operator ()
  {
    { T t = { T() }; static_assert(std::is_same<decltype(*t),U>::value,""); }
    { T const t = { T() }; static_assert(std::is_same<decltype(*t),U>::value,""); }
  }
  static void pointer_operator ()
  {
    { T t = { T() }; static_assert(std::is_same<decltype(t->f()),U>::value,""); }
    { T const t = { T() }; static_assert(std::is_same<decltype(t->f()),U>::value,""); }
  }
  
private:
  template < typename V >
  static void default_construct (V) { T t = nullptr; assert(!t); }
  template < typename V >
  static void default_construct (observer_ptr<V>) { T t; assert(!t); }
};

typedef boost::mpl::vector<int,C> test_types;

BOOST_AUTO_TEST_CASE_TEMPLATE(test_typedef,T,test_types)
{
  using raw_pointer = T*;
  using const_raw_pointer = T const*;
  using pointer = observer_ptr<T>;
  using const_pointer = observer_ptr<T const>;
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_default_constructor,T,test_types)
{
  test<T*>::default_construct();
  test<observer_ptr<T>>::default_construct();
  test<T const*>::default_construct();
  test<observer_ptr<T const>>::default_construct();
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_construct_from_mutable,T,test_types)
{
  T t;
  test<T*>::construct(&t);
  test<observer_ptr<T>>::construct(&t);
  test<T const*>::construct(&t);
  test<observer_ptr<T const>>::construct(&t);
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_construct_from_const,T,test_types)
{
  T const t = { T() };
  test<T const*>::construct(&t);
  test<observer_ptr<T const>>::construct(&t);
#if XXX_CANNOT_CONSTRUCT_MUTABLE_POINTER_FROM_CONST_POINTER
  test<T*>::construct(&t);
  test<observer_ptr<T>>::construct(&t);
#endif // XXX_CANNOT_CONSTRUCT_MUTABLE_POINTER_FROM_CONST_POINTER
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_construct_from_unique_ptr,T,test_types)
{
  using std::make_unique;
  auto pointer = make_unique<T>();
  test<observer_ptr<T>>::construct(pointer);
  test<observer_ptr<T const>>::construct(pointer);
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_initialize_from_mutable,T,test_types)
{
  T t;
  test<T*>::initialize(&t);
  test<observer_ptr<T>>::initialize(&t);
  test<T const*>::initialize(&t);
  test<observer_ptr<T const>>::initialize(&t);
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_initialize_from_const,T,test_types)
{
  T const t = { T() };
  test<T const*>::initialize(&t);
  test<observer_ptr<T const>>::initialize(&t);
#if XXX_CANNOT_CONSTRUCT_MUTABLE_POINTER_FROM_CONST_POINTER
  test<T*>::initialize(&t);
  test<observer_ptr<T>>::initialize(&t);
#endif // XXX_CANNOT_CONSTRUCT_MUTABLE_POINTER_FROM_CONST_POINTER
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_copy_construct,T,test_types)
{
  T t;
  test<T*,T*>::copy_construct(&t);
  test<observer_ptr<T>,observer_ptr<T>>::copy_construct(&t);
  test<observer_ptr<T>,T*>::copy_construct(&t);
#if XXX_CANNOT_CONSTRUCT_MUTABLE_POINTER_FROM_CONST_POINTER
  test<T*,T const*>::copy_construct(&t);
  test<observer_ptr<T>,observer_ptr<T const>>::copy_construct(&t);
  test<observer_ptr<T>,T const*>::copy_construct(&t);
#endif // XXX_CANNOT_CONSTRUCT_MUTABLE_POINTER_FROM_CONST_POINTER
  test<T const*,T*>::copy_construct(&t);
  test<observer_ptr<T const>,observer_ptr<T>>::copy_construct(&t);
  test<observer_ptr<T const>,T*>::copy_construct(&t);
  test<T const*,T const*>::copy_construct(&t);
  test<observer_ptr<T const>,observer_ptr<T const>>::copy_construct(&t);
  test<observer_ptr<T const>,T* const>::copy_construct(&t);
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_move_construct,T,test_types)
{
  T t;
  test<T*,T*>::move_construct(&t);
  test<observer_ptr<T>,observer_ptr<T>>::move_construct(&t);
  test<observer_ptr<T>,T*>::move_construct(&t);
#if XXX_CANNOT_CONSTRUCT_MUTABLE_POINTER_FROM_CONST_POINTER
  test<T*,T const*>::move_construct(&t);
  test<observer_ptr<T>,observer_ptr<T const>>::move_construct(&t);
  test<observer_ptr<T>,T const*>::move_construct(&t);
#endif // XXX_CANNOT_CONSTRUCT_MUTABLE_POINTER_FROM_CONST_POINTER
  test<T const*,T*>::move_construct(&t);
  test<observer_ptr<T const>,observer_ptr<T>>::move_construct(&t);
  test<observer_ptr<T const>,T*>::move_construct(&t);
  test<T const*,T const*>::move_construct(&t);
  test<observer_ptr<T const>,observer_ptr<T const>>::move_construct(&t);
  test<observer_ptr<T const>,T const*>::move_construct(&t);
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_copy_assign,T,test_types)
{
  T t;
  test<T*,T*>::copy_assign(&t);
  test<observer_ptr<T>,observer_ptr<T>>::copy_assign(&t);
  test<observer_ptr<T>,T*>::copy_assign(&t);
#if XXX_CANNOT_CONSTRUCT_MUTABLE_POINTER_FROM_CONST_POINTER
  test<T*,T const*>::copy_assign(&t);
  test<observer_ptr<T>,observer_ptr<T const>>::copy_assign(&t);
  test<observer_ptr<T>,T const*>::copy_assign(&t);
#endif // XXX_CANNOT_CONSTRUCT_MUTABLE_POINTER_FROM_CONST_POINTER
  test<T const*,T*>::copy_assign(&t);
  test<observer_ptr<T const>,observer_ptr<T>>::copy_assign(&t);
  test<observer_ptr<T const>,T*>::copy_assign(&t);
  test<T const*,T const*>::copy_assign(&t);
  test<observer_ptr<T const>,observer_ptr<T const>>::copy_assign(&t);
  test<observer_ptr<T const>,T const*>::copy_assign(&t);
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_move_assign,T,test_types)
{
  T t;
  test<T*,T*>::move_assign(&t);
  test<observer_ptr<T>,observer_ptr<T>>::move_assign(&t);
  test<observer_ptr<T>,T*>::move_assign(&t);
#if XXX_CANNOT_CONSTRUCT_MUTABLE_POINTER_FROM_CONST_POINTER
  test<T*,T const*>::move_assign(&t);
  test<observer_ptr<T>,observer_ptr<T const>>::move_assign(&t);
  test<observer_ptr<T>,T const*>::move_assign(&t);
#endif // XXX_CANNOT_CONSTRUCT_MUTABLE_POINTER_FROM_CONST_POINTER
  test<T const*,T*>::move_assign(&t);
  test<observer_ptr<T const>,observer_ptr<T>>::move_assign(&t);
  test<observer_ptr<T const>,T*>::move_assign(&t);
  test<T const*,T const*>::move_assign(&t);
  test<observer_ptr<T const>,observer_ptr<T const>>::move_assign(&t);
  test<observer_ptr<T const>,T const*>::move_assign(&t);
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_make_observer,T,test_types)
{
  using std::make_unique;

  T t;
  T const u = { T() };
  auto pointer = make_unique<T>();

  test<observer_ptr<T>>::test_make_observer(t);
  test<observer_ptr<T>>::test_make_observer(pointer);
  test<observer_ptr<T const>>::test_make_observer(t);
#if XXX_CANNOT_CONSTRUCT_MUTABLE_POINTER_FROM_CONST_POINTER
  test<observer_ptr<T>>::test_make_observer(u);
#endif // XXX_CANNOT_CONSTRUCT_MUTABLE_POINTER_FROM_CONST_POINTER
  test<observer_ptr<T const>>::test_make_observer(u);
  test<observer_ptr<T const>>::test_make_observer(pointer);
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_swap,T,test_types)
{
  test<T*,T*>::test_swap();
  test<observer_ptr<T>,observer_ptr<T>>::test_swap();
  test<observer_ptr<T>,T*>::test_swap();
  test<T*,observer_ptr<T>>::test_swap();
  test<T const*,T const*>::test_swap();
  test<observer_ptr<T const>,observer_ptr<T const>>::test_swap();
  test<observer_ptr<T const>,T const*>::test_swap();
  test<T const*,observer_ptr<T const>>::test_swap();
#if XXX_CANNOT_CONSTRUCT_MUTABLE_POINTER_FROM_CONST_POINTER
  test<T*,T const*>::test_swap();
  test<observer_ptr<T>,observer_ptr<T const>>::test_swap();
  test<T const*,T*>::test_swap();
  test<observer_ptr<T const>,observer_ptr<T>>::test_swap();
#endif // XXX_CANNOT_CONSTRUCT_MUTABLE_POINTER_FROM_CONST_POINTER
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_dereference_operator,T,test_types)
{
  test<T*,T&>::dereference_operator();
  test<observer_ptr<T>,T&>::dereference_operator();
  test<T const*,T const&>::dereference_operator();
  test<observer_ptr<T const>,T const&>::dereference_operator();
}

BOOST_AUTO_TEST_CASE(test_pointer)
{
  test<C*,mutable_this>::pointer_operator();
  test<observer_ptr<C>,mutable_this>::pointer_operator();
  test<C const*,const_this>::pointer_operator();
  test<observer_ptr<C const>,const_this>::pointer_operator();
}
